#!/bin/bash
# Copyright (C) 2000-2019 Synology Inc. All rights reserved.
#
# Created on:  2019-07-22

PROGNAME=$(basename "$0")
SYNOPKG="/usr/syno/bin/synopkg"
SYNOPKG_PKGNAME="DirectoryServerForWindowsDomain"
SYNOPKG_PKGDEST="/var/packages/${SYNOPKG_PKGNAME}/target"

logger_info()
{
    logger -p user.info -t "${SYNOPKG_PKGNAME}" "[HA-healthcheck] $*"
}

logger_err()
{
    logger -s -p user.err -t "${SYNOPKG_PKGNAME}" "[HA-healthcheck] $*"
}

usage()
{
    printf "usage: %s [-h|--help] [--check-processes]\\n" "$PROGNAME"
    printf "\\n"
    printf "options:\\n"
    printf "\\t--check-processes    check samba processes health\\n"
    printf "\\t-h, --help           print usage message\\n"
}

service_is_enabled()
{
    local synoads_status

    if ! synoads_status=$(${SYNOPKG} status "${SYNOPKG_PKGNAME}"); then
        return 1
    fi
    if [[ ${synoads_status} != *" is started"* ]]; then
        return 1
    fi
    if [[ ! -f "${SYNOPKG_PKGDEST}/private/sam.ldb" ]]; then
        return 1
    fi
    if [[ ! -f "${SYNOPKG_PKGDEST}/private/synoadserver.conf.hash" ]]; then
        return 1
    fi
    return 0
}

probe_services()
{
    # 88 kerberos
    # 135 dcerpc
    # 389 ldap
    # 464 krbpasswd
    # 636 ldaps (optional)
    # 1024 rpc
    # 3268 gc
    # 3269 gc/ssl (optional)
    local -r ports=(88 135 389 464 1024 3268)
    local -A check_port
    local daemon_dead=false
    local netstat_result

    if ! netstat_result=$(netstat -ntpl4 | grep samba); then
        logger_info "netstat failed. skip probing..."
        exit 0
    fi

    while IFS=$'\n' read -ra line; do
        for port in "${ports[@]}"; do
            if [[ ${line} == *":${port} "* ]]; then
                logger_info "${port} port is open"
                check_port["${port}"]=true
            fi
        done
    done <<< "${netstat_result}"

    for port in "${ports[@]}"; do
        if [[ ${check_port["${port}"]} != true ]]; then
            logger_err "samba doesn't listen on ${port} port"
            daemon_dead=true
        fi
    done

    if [[ ${daemon_dead} == true ]]; then
        logger_err "need to failover"
        exit 1
    fi
}

check_processes()
{
    if ! service_is_enabled; then
        exit 0
    fi

    service_status=$(/sbin/status "pkg-${SYNOPKG_PKGNAME}")
    case ${service_status} in
        *"start/running"*)
            probe_services
            ;;
        *)
            logger_info "service is in middle status. skip probing..."
            exit 2
            ;;
    esac
}

case $1 in
    --check-processes)
        check_processes
        ;;
    -h|--help)
        usage
        ;;
    *)
        usage >&2
        exit 2
        ;;
esac

exit 0
#vim:set sts=4 sw=4 ts=4 et:
