#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# forget_password.py
# Copyright (C) 2000-2018 Synology Inc. All rights reserved.
#
# Created by:   lukewu
# Created on:   2018-11-14

"""
the hook to change registered forgetpassword service ip addresses in synoads
"""
import os
import sys
import syslog
import subprocess
import argparse
import json
import ConfigParser


def successfully_exit(mesg):
    if len(mesg):
        syslog.syslog(
            syslog.LOG_DEBUG,
            '[DirectoryServerForWindowsDomain:forget_password] {0}'.format(mesg))
    sys.exit(0)


class ExecuterResult(object):
    def __init__(self, retcode, out, err):
        self.retcode = retcode
        self.out = out
        self.err = err

    def __nonzero__(self):
        return self.retcode == 0

    def get_toperr(self):
        return self.err[0] if len(self.err) > 0 else '(unknown)'


class Executer(object):
    def run(self, argv, strip_space=True):
        p = subprocess.Popen(argv,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)

        out, err = p.communicate()
        if strip_space:
            out = [line.strip(' \t') for line in out.splitlines() if line]
            err = [line.strip(' \t') for line in err.splitlines() if line]
        else:
            out = [line for line in out.splitlines() if line]
            err = [line for line in err.splitlines() if line]
        return ExecuterResult(p.returncode, out, err)


class SectionlessConfigParser(ConfigParser.SafeConfigParser, object):
    def __init__(self):
        super(SectionlessConfigParser, self).__init__()

    def readfp(self, fp):
        super(SectionlessConfigParser, self).readfp(
                self.SectionlessINIWrapper(fp))

    def items(self):
        config = super(SectionlessConfigParser, self).items(
                self.SectionlessINIWrapper.section)
        return [(el[0], self._normalize_value(el[1])) for el in config]

    def _normalize_value(self, value):
        value = value.strip(' \t\n\r')
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        return value

    class SectionlessINIWrapper(object):
        section = 'general'

        def __init__(self, fp):
            self.fp = fp
            self.first_time = True

        def readline(self):
            if self.first_time:
                self.first_time = False
                return '[{0}]\n'.format(self.section)
            return self.fp.readline()


class UrlHandler(object):
    pkg_root = '/var/packages/DirectoryServerForWindowsDomain'
    samldb_path = pkg_root + '/target/private/sam.ldb'
    mstch_ldif_file = pkg_root + '/conf/ldif/forget_password.ldif.mustache'
    mstch_hash_file = '/tmp/forget_password.ldif.json'
    ldif_file = '/tmp/forget_password.ldif'

    def __init__(self):
        self.executer = Executer()

    def add_service_url(self):
        mstch_hash = self._create_mstch_hash('ldbadd')
        self._do_ldb_action('ldbadd', mstch_hash)

    def update_service_url(self):
        mstch_hash = self._create_mstch_hash('ldbmodify')
        self._do_ldb_action('ldbmodify', mstch_hash)

    def _do_ldb_action(self, act_type, mstch_hash):
        with open(self.mstch_hash_file, 'w') as fp:
            json.dump(mstch_hash, fp)
        res = self.executer.run([
            '/usr/syno/bin/synomustache', '-o', self.ldif_file,
            self.mstch_ldif_file, self.mstch_hash_file])
        if not res:
            raise RuntimeError(
                'failed to create ldif file: "{0}"'.format(res.get_toperr()))

        res = self.executer.run([
            ('/bin/' + act_type), '-H', self.samldb_path, self.ldif_file])
        if not res:
            raise RuntimeError('failed to {0} ipservices: "{1}"'.format(
                act_type, res.get_toperr()))

    def _create_mstch_hash(self, act_type):
        mstch_hash = {
            'domain_dn': self._get_domain_dn(),
            act_type: [],
            'ip_list': [],
        }

        ext_url_list = self._get_external_url()

        ports = self._get_http_https_ports()
        for key in ports:
            mstch_hash[act_type].append({
                'protocol': key,
                'protocol_l': key.lower(),
                'port': ports[key],
                'ext_url': [url for url in ext_url_list
                            if url.startswith(key.lower() + '://')]
            })

        ip_list = self._get_host_ip()
        for ip in ip_list:
            mstch_hash['ip_list'].append({'ip': ip})

        return mstch_hash

    def _get_domain_dn(self):
        if not os.path.isfile(self.samldb_path):
            raise RuntimeError('no sam ldb: {0}'.format(self.samldb_path))
        res = self.executer.run([
            '/bin/ldbsearch', '-H', self.samldb_path,
            '(objectclass=domain)', 'distinguishedname'])
        if not res:
            raise RuntimeError(
                    'failed to get domain dn: {0}'.format(res.get_toperr()))

        dn_lines = [s[len('dn:'):] for s in res.out if s.startswith('dn:')]
        if not dn_lines:
            raise RuntimeError('failed to find domain dn')
        return dn_lines[0]

    def _get_http_https_ports(self):
        ports = {}
        w3_path = '/usr/syno/bin/synow3'
        w3_http = self.executer.run([w3_path, '--get-dsm-port'])
        w3_https = self.executer.run([w3_path, '--get-dsm-ssl-port'])
        http_port = os.environ.get('HTTP_PORT',
                                   int(w3_http.out[0]) if w3_http else None)
        https_port = os.environ.get('HTTPS_PORT',
                                    int(w3_https.out[0]) if w3_https else None)
        if http_port:
            ports['Http'] = http_port
        if https_port:
            ports['Https'] = https_port
        return ports

    def _get_host_ip(self):
        ip_list = []
        synonet = self.executer.run(['/usr/syno/sbin/synonet', '--show_ha'])
        if synonet:
            ip_list.extend([s[len('IP:'):].strip(' \t')
                           for s in synonet.out if s.startswith('IP:')])
        ddns = self.executer.run([
            '/usr/syno/sbin/synoddnsinfo', '--get-hostname'])
        # synoddnsinfo returns how many records are set
        if ddns.retcode > 0:
            # synoddnsinfo shows multiple hosts in a line with space-separated
            ip_list.extend([host for host in ddns.out[0].split(' ') if host])
        return ip_list

    def _get_external_url(self):
        url = []
        config = {}
        with open('/etc/synoinfo.conf', 'r') as fp:
            config_parser = SectionlessConfigParser()
            config_parser.readfp(fp)
            config = dict(config_parser.items())
        if 'external_host_ip' not in config or not config['external_host_ip']:
            return []
        for protocol in ['http', 'https']:
            port_key = 'external_port_dsm_{0}'.format(protocol)
            if port_key in config and config[port_key]:
                url.append('{protocol}://{host}:{port}'.format(
                    protocol=protocol,
                    host=config['external_host_ip'],
                    port=config[port_key]))
            else:
                url.append('{protocol}://{host}'.format(
                    protocol=protocol,
                    host=config['external_host_ip']))
        return url


# this is an interface but i'm lazy to use abc
class PluginHandler(object):
    def __init__(self):
        self.raw_config = {}

        with open('/var/packages/{0}/INFO'.format(self.package_id), 'r') as fp:
            config_parser = SectionlessConfigParser()
            config_parser.readfp(fp)
            self.raw_config = dict(config_parser.items())
        self.info = {
            'version': "1.0",
            'vendor': self.raw_config['maintainer'],
            'name': self.raw_config['package'],
            'pkgversion': self.raw_config['version'],
        }

    def action_pre(self):
        pass

    def action_post(self):
        pass

    def need_skip(self):
        return True

    def act(self, args):
        for key in self.info:
            if args[key] is True:
                print(self.info[key])
                return

        if self.need_skip():
            successfully_exit()

        for key in args:
            if args[key] and hasattr(self, 'action_' + key):
                return getattr(self, 'action_' + key)()

        raise RuntimeError('unknown arguments: {0}'.format(args))


class ForgetPasswordPluginHandler(PluginHandler):
    package_id = 'DirectoryServerForWindowsDomain'

    def __init__(self):
        super(ForgetPasswordPluginHandler, self).__init__()
        self.url_handler = UrlHandler()

    def need_skip(self):
        if not os.path.isfile(self.url_handler.samldb_path):
            successfully_exit('there is no domain controller, skip...')

    def action_post(self):
        self.url_handler.update_service_url()

    def action_add_url(self):
        self.url_handler.add_service_url()

    def action_update_url(self):
        self.url_handler.update_service_url()


def main(args):
    plugin_handler = ForgetPasswordPluginHandler()
    plugin_handler.act(vars(args))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
            description='update forget password service of synoads')
    exclusive_group = parser.add_mutually_exclusive_group(required=True)
    exclusive_group.add_argument('--name',
                                 action='store_true',
                                 help='get name')
    exclusive_group.add_argument('--version', '--sdk-mod-ver', '-v',
                                 action='store_true',
                                 help='get version')
    exclusive_group.add_argument('--pkgversion', '--pkg-ver',
                                 action='store_true',
                                 help='get pacakge version')
    exclusive_group.add_argument('--vendor',
                                 action='store_true',
                                 help='get vendor')
    exclusive_group.add_argument('--pre',
                                 action='store_true',
                                 help='do pre action')
    exclusive_group.add_argument('--post',
                                 action='store_true',
                                 help='do post action')
    exclusive_group.add_argument('--add-url', action='store_true')
    exclusive_group.add_argument('--update-url', action='store_true')
    args = parser.parse_args()
    try:
        main(args)
    except Exception as e:
        syslog.syslog(syslog.LOG_ERR, "Directory Service For Windows Domain "
                                      "forget password throw: {0}".format(e))
        raise

# vim:set shiftwidth=4 tabstop=4 expandtab:
